;;; -*- Mode: Lisp; Package: EDITOR -*-
;;;
;;; This file is part of xyzzy.
;;;

(provide "app-menu")

(in-package "editor")

(export '(*app-menu* *app-popup-menu* *menu-history-max*
	  *app-rectangle-popup-menu* *clipboard-char-encoding-popup-menu*
	  *dictionary-popup-menu* *paste-rectangle-popup-menu*
	  *command-bar-popup-menu*
	  make-char-encoding-popup-menu
	  add-file-history-to-menu *menu-display-length*
	  *init-app-menus-hook*))

(defvar *app-menu* nil)
(defvar *app-popup-menu* nil)
(defvar *app-rectangle-popup-menu* nil)
(defvar *clipboard-char-encoding-popup-menu* nil)
(defvar *dictionary-popup-menu* nil)
(defvar *paste-rectangle-popup-menu* nil)
(defvar *command-bar-popup-menu* nil)

(defvar *recent-file-popup-menu* nil)
(defvar *recent-session-popup-menu* nil)

(defvar *menu-history-max* 9)

(defvar *init-app-menus-hook* nil)

(defun init-app-menus ()

  (setq *clipboard-char-encoding-popup-menu*
	(make-char-encoding-popup-menu
	 #'(lambda (encoding)
	     (interactive)
	     (continue-pre-selection)
	     (setq *clipboard-char-encoding* encoding))
	 #'(lambda (encoding)
	     (and (eq *clipboard-char-encoding* encoding) :check))
	 *clipboard-char-encoding-list*))

  (setq *dictionary-popup-menu*
	(define-popup-menu ':dictionary-popup
	  (:item :tag-command "&Japenese-English"
	   'lookup-j2e-dictionary-selection)
	  (:item :tag-command "&Reading"
	   'lookup-reading-dictionary-selection)
	  (:item :tag-command "&English-Japenese"
	   'lookup-e2j-dictionary-selection)
	  (:item :tag-command "&Idiom"
	   'lookup-idiom-dictionary-selection)
	  (:sep :above-edict-mode)
	  (:item :tag-command "&Dictionary Mode"
	   'edict-mode
	   #'(lambda () (if edict-mode :check)))))

  (setq *paste-rectangle-popup-menu*
	(define-popup-menu ':paste-rectangle-popup
	  (:item :tag-command "&Insert"
	   'yank-rectangle
	   #'(lambda () (unless *rectangle-kill-buffer* :disable)))
	  (:item :tag-command "&OverWrite"
	   'overwrite-rectangle
	   #'(lambda () (unless *rectangle-kill-buffer* :disable)))
	  (:item :tag-command "&Apped EOL"
	   'append-rectangle
	   #'(lambda () (unless *rectangle-kill-buffer* :disable)))
	  (:item :tag-command "Insert as &Region"
	   'yank-rectangle-as-region
	   #'(lambda () (unless *rectangle-kill-buffer* :disable)))))

  (setq *app-popup-menu*
	(define-popup-menu
	  (:item :tag-command "Cu&t"
	   'kill-selection-to-clipboard :modify-any-selection)
	  (:item :tag-command "&Copy"
	   'copy-selection-to-clipboard :any-selection)
	  (:item :tag-command "&Paste"
	   'paste-from-clipboard :clipboard)
	  (:item :tag-command "Paste - rectan&gle"
	   'paste-rectangle-from-clipboard :clipboard)
	  (:menu nil "&Encoding"
	   *clipboard-char-encoding-popup-menu*)
	  (:item :tag-command "&Del"
	   'delete-selection :modify-any-selection)
	  (:item :tag-command "Select &All"
	   'selection-whole-buffer)
	  (:sep :above-dictionary-popup-menu)
	  (:menu nil "D&ictionary"
	   *dictionary-popup-menu*)
	  (:menu nil "Paste &Rectangle"
	   *paste-rectangle-popup-menu*)))

  (setq *app-rectangle-popup-menu*
	(define-popup-menu
	  (:item :tag-command "Cu&t"
	   'kill-rectangle-selection :modify-rectangle)
	  (:item :tag-command "&Copy"
	   'copy-rectangle-selection :rectangle)
	  (:menu nil "&Paste"
	   *paste-rectangle-popup-menu*)
	  (:item :tag-command "&Del"
	   'delete-rectangle-selection :modify-rectangle)
	  (:item :tag-command "Clea&r"
	   'clear-rectangle-selection :modify-rectangle)
	  (:item :tag-command "&Open"
	   'open-rectangle-selection :modify-rectangle)
	  (:item :tag-command "&String"
	   'string-rectangle-selection :modify-rectangle)))

  (or *command-bar-popup-menu*
      (setq *command-bar-popup-menu* (create-popup-menu ':command-bar-popup)))

  (or *recent-file-popup-menu*
      (setq *recent-file-popup-menu* (create-popup-menu ':recent-file-popup)))
  (or *recent-session-popup-menu*
      (setq *recent-session-popup-menu* (create-popup-menu ':recent-session-popup)))

  (setq *app-menu*
	(define-menu
	  (:popup 'file "&File"
	   (:item :tag-command "&New"
	    'new-file)
	   (:item :tag-command "&Open"
	    'open-file-dialog)
	   (:item :tag-command "&Close"
	    'close-selected-buffer
	    #'(lambda ()
		(if (= (count-buffers) 1) :disable)))
	   (:item :tag-command "&Save"
	    'save-buffer-dialog :modified)
	   (:item :tag-command "Save &As"
	    'save-as-dialog)
	   (:item :tag-command "&Insert File"
	    'insert-file-dialog)
	   (:item :tag-command "&Rename"
	    'rename-dialog)
	   (:item :tag-command "Sa&ve All"
	    'save-all-buffers)
	   (:sep :above-session)
	   (:item :tag-command "Open Session (&D)"
	    'open-session-dialog)
	   (:item :tag-command "Save Session (&W)"
	    'save-session-dialog)
	   (:item :tag-command "&Toggle session auto save"
	    'toggle-session-auto-save
	    'update-session-auto-save-ui)
	   (:item :tag-command "Close session (&Y)"
	    'close-session-dialog)
	   (:sep :above-print)
	   (:item :tag-command "&Print"
	    'print-selected-buffer-dialog)
	   (:sep :above-recent)
	   (:menu nil "Recent &File"
	    *recent-file-popup-menu*)
	   (:menu nil "Recent Session (&L)"
	    *recent-session-popup-menu*)
	   (:sep :above-kill-xyzzy)
	   (:item :tag-command "Save and &Exit"
	    'save-all-buffers-kill-xyzzy)
	   (:item :tag-command "E&xit xyzzy"
	    'kill-xyzzy))
	  (:popup 'edit "&Edit"
	   (:item :tag-command "&Undo"
	    'undo :undo)
	   (:item :tag-command "&Redo"
	    'redo :redo)
	   (:sep :above-selection)
	   (:item :tag-command "Cu&t"
	    'kill-selection-to-clipboard :modify-any-selection)
	   (:item :tag-command "&Copy"
	    'copy-selection-to-clipboard :any-selection)
	   (:item :tag-command "&Paste"
	    'paste-from-clipboard :clipboard)
	   (:item :tag-command "Paste - rectan&gle"
	    'paste-rectangle-from-clipboard :clipboard)
	   (:menu nil "&Encoding"
	    *clipboard-char-encoding-popup-menu*)
	   (:item :tag-command "De&l"
	    'delete-selection :modify-any-selection)
	   (:item :tag-command "&Select All"
	    'selection-whole-buffer)
	   (:sep :above-rectangle)
	   (:menu nil "Recta&ngle"
	    *app-rectangle-popup-menu*)
	   (:popup :convert-popup "Co&nvert"
	    (:item :tag-command "&Upcase"
	     'upcase-selection :selection)
	    (:item :tag-command "&Lowercase"
	     'downcase-selection :selection)
	    (:item :tag-command "&Capitalize"
	     'capitalize-selection :selection)
	    (:sep :above-space-tab)
	    (:item :tag-command "Space to &Tab"
	     'tabify-selection :selection)
	    (:item :tag-command "Tab to &Space"
	     'untabify-selection :selection)
	    (:sep :above-full->half)
	    (:popup :full->half-popup "Full Width to &Half Width"
	     (:item :tag-command "&All"
	      'map-all-to-half-width-selection :selection)
	     (:item :tag-command "ascii,hiragana,katakana(&S)"
	      'map-to-half-width-selection :selection)
	     (:item :tag-command "ascii,hiragana(&R)"
	      'map-ascii-and-hiragana-to-half-width-selection :selection)
	     (:item :tag-command "ascii,katakana(&T)"
	      'map-ascii-and-katakana-to-half-width-selection :selection)
	     (:item :tag-command "Ascii(&N)"
	      'map-ascii-to-half-width-selection :selection)
	     (:item :tag-command "&Hiragana"
	      'map-hiragana-to-half-width-selection :selection)
	     (:item :tag-command "&Katakana"
	      'map-katakana-to-half-width-selection :selection)
	     (:item :tag-command "&Greek"
	      'map-greek-to-half-width-selection :selection)
	     (:item :tag-command "&Cyrillic"
	      'map-cyrillic-to-half-width-selection :selection))
	    (:popup :half->full-popup "Half Width to Full Width (&Z)"
	     (:item :tag-command "ALL (To hiragana)(&A)"
	      'map-all-to-full-width-hiragana-selection :selection)
	     (:item :tag-command "ALL (To katakana)(&L)"
	      'map-all-to-full-width-katakana-selection :selection)
	     (:item :tag-command "ascii,kana (To hiragana)(&R)"
	      'map-to-full-width-hiragana-selection :selection)
	     (:item :tag-command "ALL (To full katakana)(&T)"
	      'map-to-full-width-katakana-selection :selection)
	     (:item :tag-command "Ascii(&N)"
	      'map-ascii-to-full-width-selection :selection)
	     (:item :tag-command "kana to hiragana(&H)"
	      'map-kana-to-full-width-hiragana-selection :selection)
	     (:item :tag-command "kana to katakana(&K)"
	      'map-kana-to-full-width-katakana-selection :selection)
	     (:item :tag-command "&Greek"
	      'map-greek-to-full-width-selection :selection)
	     (:item :tag-command "&Cyrillic"
	      'map-cyrillic-to-full-width-selection :selection))
	    (:popup :char-encoding-popup "&Kanji Code"
	     (:item :tag-command "&Auto"
	      'map-char-encoding-selection :selection)
	     (:item :tag-command "From &JIS"
	      'map-jis-selection :selection)
	     (:item :tag-command "From &EUC"
	      'map-euc-selection :selection)))
	   (:popup :paragraph-popup "Paragaraph(&H)"
	    (:item :tag-command "&Backward"
	     'backward-paragraph)
	    (:item :tag-command "&Forward"
	     'forward-paragraph)
	    (:item :tag-command "&Select"
	     'selection-paragraph)
	    (:item :tag-command "De&l"
	     'kill-paragraph)
	    (:item :tag-command "&Transpose"
	     'transpose-paragraphs)
	    (:item :tag-command "F&ill"
	     'fill-paragraph))
	   (:sep :above-not-modified)
	   (:item :tag-command "&Mark buffer as unmodified"
	    'not-modified :modified)
	   (:item :tag-command "&Omit changes"
	    'revert-buffer))
	  (:popup 'search "&Search"
	   (:item :tag-command "&Find"
	    'search-dialog)
	   (:item :tag-command "&Next"
	    'repeat-forward-search
	    #'(lambda () (or *last-search-p* :disable)))
	   (:item :tag-command "&Previous"
	    'repeat-backward-search
	    #'(lambda () (or *last-search-p* :disable)))
	   (:item :tag-command "&Replace"
	    'replace-dialog)
	   (:item :tag-command "&Count Matchs"
	    'count-matches-dialog)
	   (:sep :above-first-error)
	   (:item :tag-command "F&irst Error"
	    'first-error)
	   (:item :tag-command "N&ext Error"
	    'next-error)
	   (:sep :above-goto-line)
	   (:item :tag-command "&Jump to line ..."
	    'goto-line-dialog)
	   (:item :tag-command "&Top of file"
	    'beginning-of-buffer)
	   (:item :tag-command "&Bottom of file"
	    'end-of-buffer)
	   (:item :tag-command "&Mark"
	    'mark-dialog-box)
	   (:item :tag-command "beginning of function (&D)"
	    'beginning-of-defun)
	   (:item :tag-command "end of function (&U)"
	    'end-of-defun)
	   (:sep :above-gresreg)
	   (:item :tag-command "Gresreg(&S)..."
	    'gresreg-dialog)
	   (:item :tag-command "Grep(&G)..."
	    'grep-dialog))
	  (:popup 'view "&View"
	   (:item :tag-command "&No Fold"
	    'set-buffer-fold-type-none
	    'set-buffer-fold-type-none-update)
	   (:item :tag-command "Fold at set &Columns"
	    'set-buffer-fold-type-column
	    'set-buffer-fold-type-column-update)
	   (:item :tag-command "Fold at &Window width"
	    'set-buffer-fold-type-window
	    'set-buffer-fold-type-window-update)
	   (:sep :above-command-bar)
	   (:menu nil "Tool Bar (&Q)" *command-bar-popup-menu*)
	   (:sep :above-frame)
	   (:item :tag-command "New &Frame"
	    'new-pseudo-frame 'new-pseudo-frame-menu-update)
	   (:item :tag-command "Close Current Frame (&D)"
	    'delete-pseudo-frame 'pseudo-frame-menu-update)
	   (:item :tag-command "Nex&t Frame"
	    'other-pseudo-frame 'pseudo-frame-menu-update)
	   (:item :tag-command "P&revious Frame"
	    'previous-pseudo-frame 'pseudo-frame-menu-update))
	  (:popup 'window "&Window"
	   (:item :tag-command "&Split Window Horizontally"
	    'split-window 'split-window-menu-update)
	   (:item :tag-command "Split Window &Vertically"
	    'split-window-vertically 'split-window-menu-update)
	   (:item :tag-command "&One Window"
	    'delete-other-windows 'close-window-menu-update)
	   (:item :tag-command "&Close Window"
	    'delete-window 'close-window-menu-update)
	   (:item :tag-command "&Next Window"
	    'other-window 'move-window-menu-update)
	   (:item :tag-command "&Previous Window"
	    'move-previous-window 'move-window-menu-update)
	   (:sep :above-next-xyzzy)
	   (:item :tag-command "Next xyzzy (&X)"
	    'next-xyzzy-window
	    #'(lambda () (if (< (count-xyzzy-instance) 2) :disable)))
	   (:item :tag-command "Previous xyzzy (&Y)"
	    'previous-xyzzy-window
	    #'(lambda () (if (< (count-xyzzy-instance) 2) :disable)))
	   (:item :tag-command "Select xyzzy(&Z)..."
	    'select-xyzzy
	    #'(lambda () (if (< (count-xyzzy-instance) 2) :disable)))
	   (:sep :above-select-buffer)
	   (:item :tag-command "Select &Buffer..."
	    'select-buffer))

	  (:popup 'tools "&Tools"
	   (:item :tag-command "exec console prog &Asynchronously"
	    'execute-subprocess-dialog)
	   (:item :tag-command "&Kill async program"
	    'kill-subprocess
	    #'(lambda ()
		(let ((proc (buffer-process (selected-buffer))))
		  (unless (and proc
			       (eq (process-status proc) :run))
		    :disable))))
	   (:item :tag-command "Exec Console &Program ..."
	    'pipe-command-dialog)
	   (:item :tag-command "Exec &Windows Program ..."
	    'launch-application-dialog)
	   (:item :tag-command (if (featurep :windows-nt)
				   "NT Prompt(&D)"
				 "DOS Prompt(&D)")
	    'run-console)
	   (:sep :above-kbd-macro)
	   (:item :tag-command "&Start Record keyboard macro"
	    'start-kbd-macro)
	   (:item :tag-command "&End Record keyboard macro"
	    'end-kbd-macro
	    :end-macro)
	   (:item :tag-command "E&xec Keyboard Macro"
	    'call-last-kbd-macro
	    #'(lambda () (if (null *last-kbd-macro*) :disable)))
	   (:popup :kbd-macro-popup "Key board Macro Misc (&O)"
	    (:item :tag-command "Sa&ve..."
	     'save-last-kbd-macro
	     #'(lambda () (if (null *last-kbd-macro*) :disable)))
	    (:item :tag-command "&Load..."
	     'load-kbd-macro)
	    (:item :tag-command "&Save to file..."
	     'save-kbd-macro-to-file))
	   (:sep :above-property-sheet)
	   (:item :tag-command "&Common Setting..."
	    'option-property-sheet)
	   (:item :tag-command "&Local Setting..."
	    'local-option-property-sheet)
	   (:sep :above-box-drawings)
	   (:item :tag-command "&Ruling Mode"
	    'box-drawings-mode
	    #'(lambda () (if *box-drawings-mode* :check)))
	   (:item :tag-command "&Thick Rule"
	    'box-drawings-toggle-line-type
	    #'(lambda ()
		(values (unless *box-drawings-mode*
			  :disable)
			(if (eq *box-drawings-line-type* '*box-drawings-thick-line*)
			    :check))))
	   (:sep :above-insert-date)
	   (:item :tag-command "Insert date and time (&M)..."
	    'insert-date-string)
	   (:item :tag-command "&Filer..."
	    'open-filer)
	   (:menu nil "D&ictionary"
	    *dictionary-popup-menu*)
	   (:sep :above-make-tags)
	   (:item :tag-command "Make TA&GS FILE"
	    'make-tags-file-dialog)
	   (:item :tag-command "Tag &Jump"
	    'jump-tag
	    #'(lambda () (unless (and tags-find-target tags-find-point) :disable)))
	   (:item :tag-command "Direct Tag J&ump"
	    'direct-tag-jump
	    #'(lambda () (unless (and tags-find-target tags-find-point) :disable)))
	   (:item :tag-command "&Back Tag Jump"
	    'back-tag-jump
	    #'(lambda () (if (null *global-mark-list*) :disable)))
	   (:sep :|above-(>_<)|)
	   (:popup :|(>_<)-popup| "(>&_<)"
	    (:item :tag-command "Read &Nifty Log"
	     'view-nifty-log)
	    (:item :tag-command "CALC (&T)"
	     'calc)
	    (:item :tag-command "C&alendar"
	     'calendar)
	    (:item :tag-command "Tower of &Hanoi"
	     'hanoi)
	    (:item :tag-command "&Gomoku"
	     'gomoku)
	    (:item :tag-command "&Life"
	     'life)
	    (:item :tag-command "&C Curve"
	     'c-curve)
	    (:item :tag-command "D&RAGON Curve"
	     'dragon-curve)))
	  (:popup 'help "Help (&?)"
	   (:item :tag-command "&Key Bindings List "
	    'describe-bindings)
	   (:item :tag-command "View &Lossage"
	    'view-lossage)
	   (:sep :above-about)
	   (:item :tag-command "&About ..."
	    'about-dialog))))

  (setq *buffer-bar-context-menu*
	(define-popup-menu
	  (:item 'close "&Close"
	   #'(lambda ()
	       (interactive)
	       (kill-buffer *buffer-bar-context-menu-buffer*)))))

  (run-hooks '*init-app-menus-hook*)
  (set-menu *app-menu*)
  (add-file-history-to-menu)
  (add-session-history-to-menu))

(defun make-char-encoding-popup-menu (command &optional init (list *char-encoding-list*))
  (labels ((define-menu (list)
	     (let ((menu (create-popup-menu)))
	       (mapc #'(lambda (encoding)
			 (if (consp encoding)
			     (add-popup-menu menu
					     (define-menu (cdr encoding))
					     (car encoding))
			   (add-menu-item menu nil
					  (char-encoding-display-name encoding)
					  #'(lambda ()
					      (interactive)
					      (funcall command encoding))
					  (and init #'(lambda ()
							(funcall init encoding))))))
		     list)
	       menu)))
    (define-menu
      (if (eq list *char-encoding-list*)
	*char-encoding-tree*
	(labels ((make-tree (tree list)
		   (mapcan #'(lambda (encoding)
			       (if (consp encoding)
				   (let ((l (make-tree (cdr encoding) list)))
				     (and l (list (cons (car encoding) l))))
				 (and (find encoding list :test #'eq)
				      (list encoding))))
			   tree)))
	  (make-tree
	   *char-encoding-tree*
	   (mapcar #'(lambda (encoding)
		       (if (symbolp encoding) (symbol-value encoding) encoding))
	     list)))))))
#|
(defun menu-display-name (path maxlen)
  (when (<= (length path) maxlen)
    (return-from menu-display-name path))
  (let* ((unc )
	 (paths (split-string path #\/))
	 (dev (cond ((eql (string/= path "//") 2)
		     (prog1
			 (concat "//" (car paths) "/" (cadr paths))
		       (setq paths (cddr paths))))
		    (t
		     (pop paths))))
	 (name (progn
		 (setq paths (nreverse paths))
		 (pop paths)))
	 (len (+ (length dev) (length name) 5)))
    (unless name
      (return-from menu-display-name path))
    (when (> len maxlen)
      (return-from menu-display-name name))
    (do* ((npaths nil)
	  (l (length (car paths)) (length (car paths))))
	((> (setq len (+ len l 1)) maxlen)
	 (format nil "~A/.../~{~A/~}~A" dev npaths name))
      (push (pop paths) npaths))))
|#

(defvar *menu-display-length* 40)

(defun add-lru-menu (menu files fn)
  (when (menup menu)
    (while (delete-menu menu 0 t))
    (do ((file files (cdr file))
	 (count 1))
	((or (null file)
	     (> count *menu-history-max*))
	 t)
      (unless (wild-pathname-p (car file))
	(let ((filename (namestring (car file))))
	  (insert-menu-item menu (1- count) nil
			    (format nil "&~:(~35r~): ~a"
				    (rem count 36)
				    (quote-string
				     (abbreviate-display-string
				      filename *menu-display-length* t)
				     #\& #\&))
			    (funcall fn filename)))
	(incf count)))))

(defun add-file-history-to-menu ()
  (when (menup *app-menu*)
    (add-lru-menu *recent-file-popup-menu*
		  *minibuffer-file-name-history*
		  #'(lambda (filename)
		      #'(lambda ()
			  (interactive)
			  (find-file filename)
			  (add-history filename '*minibuffer-file-name-history*))))))
